local Libra = LibStub("Libra")
local LIB = LibStub("LibItemButton")

local addon, addonName = Libra:NewAddon(...)
Libra:Embed(addon)

local Recipes = addon.recipes

local myRealm = GetRealmName()
local myFaction = UnitFactionGroup("player")
local myName = UnitName("player")
local myCharacter = myName.. " - "..myRealm
local recipeColorCache = {}
local options, lastRecipe

local ERR_SKILL_GAINED_S	= ERR_SKILL_GAINED_S:format("(.+)")		-- "You have gained the (.+) skill."
local ERR_SPELL_UNLEARNED_S	= ERR_SPELL_UNLEARNED_S:format("(.+)%")	-- "You have unlearned (.+)%."
local ERR_SKILL_UP_SI		= format("^%s$", ERR_SKILL_UP_SI:gsub("%%[sd]", "(%0+)"):gsub("%%s", "."))			-- "^Your skill in (.+) has increased to (%d+).$"

addon.menu = {}

-- "apprentice" rank spell IDs, used for localization
addon.skills = {
	2018, -- Blacksmithing
	2108, -- Leatherworking
	2259, -- Alchemy
	2550, -- Cooking
	-- 2575, -- Mining
	3273, -- First Aid
	3908, -- Tailoring
	4036, -- Engineering
	7411, -- Enchanting
	25229, -- Jewelcrafting
	45357, -- Inscription
}
local skills = addon.skills

-- this lets us lookup trade skill spell IDs by name, eg:
-- skillsByName["Blacksmithing"] = 2018
-- used for getting trade skill ID from recipes and the trade skill frame
local skillsByName = {}

for i = 1, #skills do
	local spellID = skills[i]
	local spellName = GetSpellInfo(spellID)
	skillsByName[spellName] = spellID
	skills[spellID] = spellName
	skills[i] = nil
end

local specs = {
	[20219] = 4036, -- Gnomish Engineer
	[20222] = 4036, -- Goblin Engineer
}

local specsByName = {}

for spellID in pairs(specs) do
	specsByName[GetSpellInfo(spellID)] = spellID
end


local tooltip = CreateFrame("GameTooltip", "BoCTooltip", GameTooltip, "ShoppingTooltipTemplate")
-- _G[self:GetName().."TextRight1"]:SetFontObject("GameFontNormalSmall")
BoCTooltipTextLeft2:SetFontObject("GameFontNormalSmall")
tooltip:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b)
tooltip:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b)
local addLine = tooltip.AddLine
function tooltip:AddLine(text, color)
	if color then
		addLine(self, text, color.r, color.g, color.b)
	else
		addLine(self, text, 1.0, 1.0, 1.0)
	end
end


local function onClick(self, link)
	SetItemRef(link:match("|H(.-)|h"), link)
end

local dropdown = addon:CreateDropdown("Menu")
dropdown.initialize = function(self, level)
	if level == 1 then
		local info = UIDropDownMenu_CreateInfo()
		info.text = myName
		info.value = myCharacter
		info.hasArrow = true
		info.notCheckable = true
		self:AddButton(info, level)
		for i, v in ipairs(addon.menu[myRealm].charNames) do
			if v ~= myCharacter and #addon.menu[myRealm].data[v] > 0 then
				local info = UIDropDownMenu_CreateInfo()
				info.text = v:gsub(" %- "..myRealm, "")
				info.value = v
				info.hasArrow = true
				info.notCheckable = true
				self:AddButton(info, level)
			end
		end
	elseif level == 2 then
		for k, v in pairs(addon.db.skillData[UIDROPDOWNMENU_MENU_VALUE].skills) do
			local info = UIDropDownMenu_CreateInfo()
			info.text = skills[k]
			info.func = onClick
			info.arg1 = v.link
			info.notCheckable = true
			self:AddButton(info, level)
		end
	end
end

local dataobj = LibStub("LibDataBroker-1.1"):NewDataObject(addonName, {
	type = "launcher",
	icon = [[Interface\Icons\INV_Misc_Book_09]],
	label = "Book of Crafts",
	OnClick = function(self)
		dropdown:Toggle(nil, self)
	end,
})


--[[

DB = {
	["Character - Realm"] = {
		skills = {
			[spellID] = {
				crafts = {
					-- any learnt trade skill crafts
					[craftID] = true,
					[...]
				},
				-- any learnt specialisation of the trade skill such as gnomish engineering, etc
				spec = spellID,
				link = "link",
				level = tradeSkillLevel,
			},
		},
		faction = "Faction",
	},
}

]]

local defaults = {
	options = {
		sameFaction = true,
		separateTooltip = false,
		colorizeRecipes = true,
		colors = {
			usable   = {r = 0.0, g = 1.0, b = 0.0},
			lowSkill = {r = 1.0, g = 0.0, b = 0.0},
			noSpec   = {r = 1.0, g = 1.0, b = 0.0},
			knownBy  = {r = 0.5, g = 0.5, b = 1.0},
			useless  = {r = 0.5, g = 0.5, b = 0.5},
		},
	},
	skillData = {},
}

function addon:OnInitialize()
	local db = self:CreateDB("BookofCraftsDB", defaults)
	self.db = db
	options = db.options
	
	-- clear old data
	db.data = nil
	
	db.skillData[myCharacter] = db.skillData[myCharacter] or {skills = {}}
	self.player = db.skillData[myCharacter]
	self.player.faction = myFaction
	
	self:LoadSettings()
	
	GameTooltip:HookScript("OnTooltipSetItem", self.ScanTooltip)
	GameTooltip:HookScript("OnTooltipCleared", function(self)
		self.BoC = nil
	end)
	
	self:RegisterEvent("PLAYER_LOGIN", "ScanProfessions")
	self:RegisterEvent("UNIT_NAME_UPDATE", "ScanProfessions")
	self:RegisterEvent("TRADE_SKILL_UPDATE")
	self:RegisterEvent("LEARNED_SPELL_IN_TAB")
	self:RegisterEvent("CHAT_MSG_SKILL")
	self:RegisterEvent("CHAT_MSG_SYSTEM")
end

function addon:TRADE_SKILL_UPDATE()
	if IsTradeSkillLinked() or IsTradeSkillGuild() or IsNPCCrafting() then return end
	
	-- return if not yet fully loaded
	if select(2, GetTradeSkillInfo(1)) ~= "header" then return end
	
	local tradeSkill, skillLevel = GetTradeSkillLine()
	tradeSkill = skillsByName[tradeSkill]

	if tradeSkill then
		tradeSkill = self.player.skills[tradeSkill]

		tradeSkill.link = GetTradeSkillListLink()
		tradeSkill.level = skillLevel
		local newCraft
		local crafts = tradeSkill.crafts

		for i = 1, GetNumTradeSkills() do
			local recipeLink = GetTradeSkillRecipeLink(i)
			
			if recipeLink then
				local craftID = tonumber(recipeLink:match("enchant:(%d+)"))
				if not crafts[craftID] then
					crafts[craftID] = true
					newCraft = true
				end
			end
		end
		if newCraft then
			self:ClearCache()
		end
	end
end

function addon:LEARNED_SPELL_IN_TAB(spellID)
	if specs[spellID] then
		self:ScanProfessions()
	end
end

function addon:CHAT_MSG_SKILL(msg)
	local tradeSkillName, newLevel = msg:match(ERR_SKILL_UP_SI)
	local tradeSkill = self.player.skills[skillsByName[tradeSkillName]]

	if tradeSkill and newLevel then
		tradeSkill.level = tonumber(newLevel)
		tradeSkill.link = select(2, GetSpellLink(tradeSkillName))
		self:ClearCache()
	end
	
	local tradeskill = msg:match(ERR_SKILL_GAINED_S)
	
	if tradeskill and skillsByName[tradeskill] then
		self:ScanProfessions()
		self:ClearCache()
	end
end

function addon:CHAT_MSG_SYSTEM(msg)
	local match = msg:match(ERR_SPELL_UNLEARNED_S)
	if match then
		local spellID = tonumber(match:match("spell:(%d+)"))
		if skills[spellID] then
			-- this spell ID matches a valid tradeskill, so we'll delete it from the database
			addon.player.skills[spellID] = nil
			addon:ScanProfessions()
			addon:ClearCache()
		else
			-- this spell ID matches a valid tradeskill spec, so we'll delete the spec entry
			local skill = specs[spellID]
			if skill then
				addon.player.skills[skill].spec = nil
				addon:ScanProfessions()
				addon:ClearCache()
			end
		end
	end
end

local function sortCharacters(a, b)
	local characterA, realmA = a:match("(.+) %- (.+)")
	local characterB, realmB = b:match("(.+) %- (.+)")
	if realmA ~= realmB then
		-- your own realm gets sorted before others
		if realmA == myRealm then
			return true
		end
		if realmB == myRealm then
			return false
		end
		return realmA < realmB
	else
		return characterA < characterB
	end
end

local sortedCharacters = {}

function addon:BuildMenu()
	wipe(self.menu)
	-- sorted-by-name array for the LDB dropdown
	local realms = {}
	for characterKey, data in pairs(self.db.skillData) do
		local character, realm = characterKey:match("(.+) %- (.+)")
		realms[realm] = realms[realm] or {
			charNames = {},
			data = {},
		}
		tinsert(realms[realm].charNames, characterKey)
		if self:IsConnectedRealm(realm) then
			tinsert(realms[realm].charNames, characterKey)
		end
		local skills = {}
		for spellID in pairs(data.skills) do
			tinsert(skills, spellID)
		end
		sort(skills)
		realms[realm].data[characterKey] = skills
	end
	local sortedRealms = {}
	for realm, data in pairs(realms) do
		sort(data.charNames, sortCharacters)
		self.menu[realm] = data
		tinsert(sortedRealms, realm)
	end
	sort(sortedRealms)
	self.menu.sortedRealms = sortedRealms
end

function addon:ScanProfessions()
	for i = 1, select("#", GetProfessions()) do
		local prof = select(i, GetProfessions())
		if prof then
			local name, _, skillLevel, _, numAbilities, spellOffset = GetProfessionInfo(prof)
			local skillID = skillsByName[name]
			if skillID then
				local skills = self.player.skills
				skills[skillID] = skills[skillID] or {crafts = {}}
				local _, link = GetSpellLink(skillID)
				-- links are not available on first PLAYER_LOGIN, in which case we wait for UNIT_NAME_UPDATE
				if not link then
					return
				end
				skills[skillID].level = skillLevel
				skills[skillID].link = link
				if numAbilities == 2 then
					-- we have two profession abilities, one of them may be a profession spec
					skills[skillID].spec = specsByName[GetSpellInfo(spellOffset + 2, BOOKTYPE_PROFESSION)]
				end
			end
		end
	end
	-- if we reached this point, the links were available and we no longer need the following event
	self:UnregisterEvent("UNIT_NAME_UPDATE")
	self:BuildMenu()
end

function addon:IsRecipe(item)
	if not item then return end
	local itemID = tonumber(item) or tonumber(item:match("item:(%d+)"))
	return Recipes[itemID]
end

local cache = {
	usable = {},
	lowSkill = {},
	known = {},
	noSpec = {},
}

function addon:GetAltsInfo(item)
	local usable, lowSkill, known, noSpec = cache.usable, cache.lowSkill, cache.known, cache.noSpec, cache.bank
	
	local itemID = tonumber(item) or tonumber(item:match("item:(%d+)"))
	local recipe = Recipes[itemID]
	
	for k, v in pairs(cache) do
		wipe(v)
	end
	
	for characterKey, characterD in pairs(self.db.skillData) do
		local character, realm = characterKey:match("(.+) %- (.+)")
		local skill = characterD.skills[recipe.skill]
		if skill and self:IsConnectedRealm(realm, true) and not (characterD.ignore or skill.ignore) and (not options.sameFaction or characterD.faction == myFaction) then
			if skill.crafts[recipe.craft] then
				-- if this recipe exists in the character's database, then the character knows it
				known[#known + 1] = character
			else
				if recipe.spec and specsByName[recipe.spec] ~= skill.spec then
					local text = character
					if options.showSkillLevel then
						text = format("%s (%d)", text, skill.level)
					end
					noSpec[#noSpec + 1] = text
				else
					-- if no spec is required, or the character has that spec, then they'll be able to learn this recipe..
					if skill.level >= recipe.level then
						-- ..now, since the character's skill level is high enough..
						usable[#usable + 1] = character
					else
						-- ..or later
						lowSkill[#lowSkill + 1] = format("%s (%d)", character, skill.level)
					end
				end
			end
		end
	end

	return usable, lowSkill, known, noSpec
end

local isUseful

local function dostuff(data, caption, color)
	if #data > 0 then
		tooltip:AddLine(caption, color)
		for i, v in ipairs(data) do
			tooltip:AddLine("  "..v, color)
		end
		isUseful = true
	end
end

local function dostuff2(tooltip, data, caption, color)
	if #data > 0 then
		if color then
			tooltip:AddLine(caption..table.concat(data, ", "), color.r, color.g, color.b)
		else
			tooltip:AddLine(caption..table.concat(data, ", "), 1.0, 1.0, 1.0)
		end
		isUseful = true
	end
end

-- add extra tooltip information if necessary
function addon:ScanTooltip()
	local _, itemLink = self:GetItem()
	if self.BoC or not addon:IsRecipe(itemLink) then
		lastRecipe = nil
		return
	end
	self.BoC = true
	
	local colors = options.colors

	local usable, lowSkill, known, noSpec = cache.usable, cache.lowSkill, cache.known, cache.noSpec

	-- if this recipe is not the same as the last scanned recipe, we need to get new data
	if itemLink ~= lastRecipe then
		usable, lowSkill, known, noSpec = addon:GetAltsInfo(itemLink)
		lastRecipe = itemLink
	end

	isUseful = false
	
	if options.separateTooltip then
		tooltip:SetOwner(self, "ANCHOR_NONE")

		dostuff(usable, "Usable:", colors.usable)
		dostuff(lowSkill, "Insufficient skill:", colors.lowSkill)
		dostuff(noSpec, "Wrong spec:", colors.noSpec)
		dostuff(known, "Already known:", colors.knownBy)
		
		if not isUseful then
			tooltip:AddLine("Unuseful recipe")
		end
		
		addon:TooltipShow(self)
	else
		self:AddLine(" ")

		dostuff2(self, usable,	"Usable: ",				colors.usable)
		dostuff2(self, lowSkill,"Insufficient skill: ",	colors.lowSkill)
		dostuff2(self, noSpec,	"Wrong spec: ",			colors.noSpec)
		dostuff2(self, known,	"Already known: ",		colors.knownBy)
		
		if not isUseful then
			self:AddLine("Unuseful recipe", 1.0, 1.0, 1.0)
		end
	end
end

function addon:TooltipShow(frame)
	local tooltipWidth = tooltip:GetWidth()
	local left = (frame:GetLeft() or 0) - tooltipWidth
	local right = GetScreenWidth() - (frame:GetRight() or 0) - tooltipWidth
	if left < right then
		tooltip:SetPoint("TOPLEFT", frame, "TOPRIGHT")
	else
		tooltip:SetPoint("TOPRIGHT", frame, "TOPLEFT")
	end

	local color = TOOLTIP_DEFAULT_BACKGROUND_COLOR
	tooltip:SetBackdropColor(color.r, color.g, color.b)
	tooltip:Show()
end

function addon:DeleteCharacterData(character, skill)
	if skill then
		self.db.skillData[character].skills[skill] = nil
		self:ScanProfessions()
	else
		self.db.skillData[character] = nil
	end
	self:ClearCache()
	self:BuildMenu()
end

function addon:ClearCache()
	lastRecipe = nil
	wipe(recipeColorCache)
	for button in pairs(LIB:GetRegisteredButtons()) do
		LIB:UpdateButton(button, LIB.buttonItems[button])
	end
	if MerchantFrame:IsShown() then
		MerchantFrame_Update()
	end
	if IsAddOnLoaded("Blizzard_AuctionUI") and AuctionFrame:IsShown() then
		AuctionFrameBrowse_Update()
	end
end

function addon:GetRecipeColor(itemLink)
	local color = recipeColorCache[itemLink]
	if not color then
		local usable, lowSkill, knownBy, noSpec = self:GetAltsInfo(itemLink)
		color = options.colors.useless
		if #noSpec > 0 then
			color = options.colors.noSpec
		elseif #usable > 0 then
			color = options.colors.usable
			if #knownBy > 0 then
				color = {r = 0, g = 1, b = 1}
			end
		elseif #knownBy > 0 then
			color = options.colors.knownBy
		elseif #lowSkill > 0 then
			color = options.colors.lowSkill
		end
		recipeColorCache[itemLink] = color
	end
	return color
end

hooksecurefunc("ContainerFrame_UpdateCooldown", function(container, button)
	local link = GetContainerItemLink(container, button:GetID())
	if options.colorizeRecipes and link and addon:IsRecipe(link) then
		local color = addon:GetRecipeColor(link)
		SetItemButtonTextureVertexColor(button, color.r, color.g, color.b)
	end
end)

local coloredCategories = {
	["MERCHANT"] = true,
	["AUCTION_BROWSE"] = true,
	["BLACKMARKET"] = true,
	["BLACKMARKET_HOT"] = true,
}

LIB:RegisterUpdateCallback(addon, function(self, button, item, category)
	if options.colorizeRecipes and item and addon:IsRecipe(item) then
		local color = addon:GetRecipeColor(item)
		if category == "MERCHANT" and MerchantFrame.selectedTab == 1 then
			local r, g, b = color.r, color.g, color.b
			local id = tonumber(strmatch(button:GetName(), "%d+"))
			local index = (MerchantFrame.page - 1) * MERCHANT_ITEMS_PER_PAGE + id
			local _, _, _, _, numAvailable, isUsable = GetMerchantItemInfo(index)
			if numAvailable == 0 then
				r, g, b = r * 0.5, g * 0.5, b * 0.5
				button.icon:SetVertexColor(r, g, b)
			else
				button.icon:SetVertexColor(r, g, b)
			end
		else
			button.icon:SetVertexColor(color.r, color.g, color.b)
		end
	elseif not coloredCategories[category] then
		button.icon:SetVertexColor(1.0, 1.0, 1.0)
	end
end)